@@ -3,7 +3,7 @@ use warnings;
 
 package HTTP::Async;
 
-our $VERSION = '0.09';
+our $VERSION = '0.33';
 
 use Carp;
 use Data::Dumper;
@@ -42,7 +42,7 @@ OR do something else if there is no response ready:
             # deal with $response
         } else {
             # do something else
-        {
+        }
     }
 
 OR just use the async object to fetch stuff in the background and deal with
@@ -81,13 +81,23 @@ C<select> lists.
 There are a number of default settings that should be suitable for most uses.
 However in some circumstances you might wish to change these.
 
-            slots: 20
-          timeout: 180 (seconds)
- max_request_time: 300 (seconds)
-    max_redirects: 7
-    poll_interval: 0.05 (seconds)
-       proxy_host: ''
-       proxy_port: ''
+            slots:  20
+          timeout:  180 (seconds)
+ max_request_time:  300 (seconds)
+    max_redirect:   7
+    poll_interval:  0.05 (seconds)
+       proxy_host:  ''
+       proxy_port:  ''
+       local_addr:  ''
+       local_port:  ''
+       ssl_options: {}
+       cookie_jar:  undef
+       peer_addr:   ''
+
+If defined, is expected to be similar to C<HTTP::Cookies>, with extract_cookies and add_cookie_header methods.
+
+The option max_redirects has been renamed to max_redirect to be consistent with LWP::UserAgent, although max_redirects still works.
+
        
 =head1 METHODS
 
@@ -106,10 +116,11 @@ sub new {
 
         opts => {
             slots            => 20,
-            max_redirects    => 7,
+            max_redirect     => 7,
             timeout          => 180,
             max_request_time => 300,
             poll_interval    => 0.05,
+            cookie_jar       => undef,
         },
 
         id_opts => {},
@@ -136,7 +147,7 @@ sub _init {
 
 sub _next_id { return ++$_[0]->{current_id} }
 
-=head2 slots, timeout, max_request_time, poll_interval, max_redirects, proxy_host and proxy_port
+=head2 slots, timeout, max_request_time, poll_interval, max_redirect, proxy_host, proxy_port, local_addr, local_port, ssl_options, cookie_jar, peer_addr
 
     $old_value = $async->slots;
     $new_value = $async->slots( $new_value );
@@ -148,11 +159,9 @@ Slots is the maximum number of parallel requests to make.
 
 =cut
 
-my %GET_SET_KEYS =
-  map { $_ => 1 }
-  qw( slots poll_interval
-  timeout max_request_time max_redirects
-  proxy_host proxy_port );
+my %GET_SET_KEYS = map { $_ => 1 } qw( slots poll_interval
+  timeout max_request_time max_redirect
+  proxy_host proxy_port local_addr local_port ssl_options cookie_jar peer_addr);
 
 sub _add_get_set_key {
     my $class = shift;
@@ -160,10 +169,15 @@ sub _add_get_set_key {
     $GET_SET_KEYS{$key} = 1;
 }
 
+my %KEY_ALIASES = ( max_redirects => 'max_redirect' );
+
 sub _get_opt {
     my $self = shift;
     my $key  = shift;
     my $id   = shift;
+
+    $key = $KEY_ALIASES{$key} if exists $KEY_ALIASES{$key};
+
     die "$key not valid for _get_opt" unless $GET_SET_KEYS{$key};
 
     # If there is an option set for this id then use that, otherwise fall back
@@ -178,6 +192,9 @@ sub _get_opt {
 sub _set_opt {
     my $self = shift;
     my $key  = shift;
+
+    $key = $KEY_ALIASES{$key} if exists $KEY_ALIASES{$key};
+
     die "$key not valid for _set_opt" unless $GET_SET_KEYS{$key};
     $self->{opts}{$key} = shift if @_;
     return $self->{opts}{$key};
@@ -225,13 +242,20 @@ This method lets you add a single request to the queue with options that
 differ from the defaults. For example you might wish to set a longer timeout
 or to use a specific proxy. Returns the id of the request.
 
+The method croaks when passed an invalid option.
+
 =cut
 
 sub add_with_opts {
     my $self = shift;
     my $req  = shift;
     my $opts = shift;
-    my $id   = $self->_next_id;
+
+    for my $key (keys %{$opts}) {
+        croak "$key not valid for add_with_opts" unless $GET_SET_KEYS{$key};
+    }
+
+    my $id = $self->_next_id;
 
     push @{ $$self{to_send} }, [ $req, $id ];
     $self->{id_opts}{$id} = $opts;
@@ -245,7 +269,7 @@ sub add_with_opts {
     $async->poke;
 
 At fairly frequent intervals some housekeeping needs to performed - such as
-reading recieved data and starting new requests. Calling C<poke> lets the
+reading received data and starting new requests. Calling C<poke> lets the
 object do this and then return quickly. Usually you will not need to use this
 as most other methods do it for you.
 
@@ -336,23 +360,58 @@ sub _next_response {
       : $resp_and_id->[0];
 }
 
-=head2 to_send_count, to_return_count, in_progress_count and total_count
+=head2 to_send_count
 
     my $pending = $async->to_send_count;
 
-Returns the number of items in the various stages of processing.
+Returns the number of items which have been added but have not yet started being processed.
 
 =cut
 
-sub to_send_count   { my $s = shift; $s->poke; scalar @{ $$s{to_send} }; }
-sub to_return_count { my $s = shift; $s->poke; scalar @{ $$s{to_return} }; }
+sub to_send_count {
+    my $self = shift;
+    $self->poke;
+    return scalar @{ $self->{to_send} };
+}
+
+=head2 to_return_count
+
+    my $completed = $async->to_return_count;
+
+Returns the number of items which have completed transferring, and are waiting to be returned by next_response().
+
+=cut
+
+sub to_return_count {
+    my $self = shift;
+    $self->poke;
+    return scalar @{ $self->{to_return} };
+}
+
+=head2 in_progress_count
+
+    my $running = $async->in_progress_count;
+
+Returns the number of items which are currently being processed asynchronously.
+
+=cut
 
 sub in_progress_count {
-    my $s = shift;
-    $s->poke;
-    scalar keys %{ $$s{in_progress} };
+    my $self = shift;
+    $self->poke;
+    return scalar keys %{ $self->{in_progress} };
 }
 
+=head2 total_count
+
+    my $total = $async->total_count;
+
+Returns the sum of the to_send_count, in_progress_count and to_return_count.
+
+This should be the total number of items which have been added that have not yet been returned by next_response().
+
+=cut
+
 sub total_count {
     my $self = shift;
 
@@ -383,6 +442,65 @@ sub info {
     );
 }
 
+=head2 remove
+
+    $async->remove($id);
+    my $success = $async->remove($id);
+
+Removes the item with the given id no matter which state it is currently in. Returns true if an item is removed, and false otherwise.
+
+=cut
+
+sub remove {
+    my $self = shift;
+    my $id = shift;
+
+    my $hashref = delete $self->{in_progress}{$id};
+    if (!$hashref) {
+        for my $list ('to_send', 'to_return') {
+            my ($r_and_id) = grep { $_->[1] eq $id } @{ $self->{$list} };
+            $hashref = $r_and_id->[0];
+            if ($hashref) {
+                @{ $self->{$list} }
+                    = grep { $_->[1] ne $id } @{ $self->{$list} };
+            }
+        }
+    }
+    return if !$hashref;
+
+    my $s = $hashref->{handle};
+    $self->_io_select->remove($s);
+    delete $self->{id_opts}{$id};
+
+    return 1;
+}
+
+=head2 remove_all
+
+    $async->remove_all;
+    my $success = $async->remove_all;
+
+Removes all items no matter what states they are currently in. Returns true if any items are removed, and false otherwise.
+
+=cut
+
+sub remove_all {
+    my $self = shift;
+    return if $self->empty;
+
+    my @ids = (
+        (map { $_->[1] } @{ $self->{to_send} }),
+        (keys %{ $self->{in_progress} }),
+        (map { $_->[1] } @{ $self->{to_return} }),
+    );
+
+    for my $id (@ids) {
+        $self->remove($id);
+    }
+
+    return 1;
+}
+
 =head2 empty, not_empty
 
     while ( $async->not_empty ) { ...; }
@@ -427,15 +545,20 @@ sub DESTROY {
 # they have been fully received yet.
 
 sub _process_in_progress {
-    my $self = shift;
+    my $self     = shift;
+    my %seen_ids = ();
 
   HANDLE:
     foreach my $s ( $self->_io_select->can_read(0) ) {
 
-        my $id = $self->{fileno_to_id}{ $s->fileno };
-        die unless $id;
+        # Get the id and add it to the hash of seen ids so we don't check it
+        # later for errors.
+        my $id = $self->{fileno_to_id}{ $s->fileno }
+          || die "INTERNAL ERROR: could not got id for fileno";
+        $seen_ids{$id}++;
+
         my $hashref = $$self{in_progress}{$id};
-        my $tmp     = $hashref->{tmp} ||= {};
+        my $tmp = $hashref->{tmp} ||= {};
 
         # warn Dumper $hashref;
 
@@ -469,10 +592,6 @@ sub _process_in_progress {
 
             # warn "Received " . length( $buf ) ;
 
-            # Reset the timeout.
-            # warn( "reseting the timeout " . time );
-            $hashref->{timeout_at} = time + $self->_get_opt( 'timeout', $id );
-
             # warn $buf;
         }
 
@@ -509,12 +628,13 @@ sub _process_in_progress {
                 my @headers_array = map { $_, $headers{$_} } keys %headers;
                 $$tmp{headers} = \@headers_array;
 
-                # Reset the timeout.
-                $hashref->{timeout_at} =
-                  time + $self->_get_opt( 'timeout', $id );
             }
         }
 
+        # Reset the timeout.
+        $hashref->{timeout_at} = time + $self->_get_opt( 'timeout', $id );
+        # warn "recieved - timeout set to '$hashref->{timeout_at}'";
+
         # If the message is complete then create a request and add it
         # to 'to_return';
         if ( $$tmp{is_complete} ) {
@@ -523,20 +643,32 @@ sub _process_in_progress {
 
             # warn Dumper $$hashref{content};
 
-            my $response =
-              HTTP::Response->new(
+            my $response = HTTP::Response->new(
                 @$tmp{ 'code', 'message', 'headers', 'content' } );
 
             $response->request( $hashref->{request} );
             $response->previous( $hashref->{previous} ) if $hashref->{previous};
 
+            # Deal with cookies
+            my $jar = $self->_get_opt('cookie_jar', $id);
+            if ($jar) {
+                $jar->extract_cookies($response);
+            }
+
             # If it was a redirect and there are still redirects left
             # create a new request and unshift it onto the 'to_send'
             # array.
+            # Only redirect GET and HEAD as per RFC  2616.
+            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
+            my $code = $response->code;
+            my $get_or_head = $response->request->method =~ m{^(?:GET|HEAD)$};
+
             if (
-                $response->is_redirect            # is a redirect
-                && $hashref->{redirects_left} > 0 # and we still want to follow
-                && $response->code != 304         # not a 'not modified' reponse
+                $response->is_redirect                     # is a redirect
+                && $hashref->{redirects_left} > 0          # and we still want to follow
+                && ($get_or_head || $code !~ m{^30[127]$}) # must be GET or HEAD if it's 301, 302 or 307
+                && $code != 304                            # not a 'not modified' reponse
+                && $code != 305                            # not a 'use proxy' response
               )
             {
 
@@ -550,7 +682,27 @@ sub _process_in_progress {
 
                 my $url = _make_url_absolute( url => $loc, ref => $uri );
 
-                my $request = HTTP::Request->new( 'GET', $url );
+                my $request = $response->request->clone;
+                $request->uri($url);
+
+                # These headers should never be forwarded
+                $request->remove_header('Host', 'Cookie');
+
+                # Don't leak private information.
+                # http://www.w3.org/Protocols/rfc2616/rfc2616-sec15.html#sec15.1.3
+                if ($request->header('Referer') &&
+                    $hashref->{request}->uri->scheme eq 'https' &&
+                    $request->uri->scheme eq 'http') {
+
+                    $request->remove_header('Referer');
+                }
+
+                # See Other should use GET
+                if ($code == 303 && !$get_or_head) {
+                    $request->method('GET');
+                    $request->content('');
+                    $request->remove_content_headers;
+                }
 
                 $self->_send_request( [ $request, $id ] );
                 $hashref->{previous} = $response;
@@ -564,6 +716,41 @@ sub _process_in_progress {
         }
     }
 
+    # warn Dumper(
+    #     {
+    #         in_progress => $self->{in_progress},
+    #         seen_ids    => \%seen_ids,
+    #     }
+    # );
+
+    foreach my $id ( keys %{ $self->{in_progress} } ) {
+
+        # skip this one if it was processed above.
+        next if $seen_ids{$id};
+
+        my $hashref = $self->{in_progress}{$id};
+
+        if (   time > $hashref->{timeout_at}
+            || time > $hashref->{finish_by} )
+        {
+
+            # warn Dumper( { hashref => $hashref, now => time } );
+
+            # we have a request that has timed out - handle it
+            $self->_add_error_response_to_return(
+                id       => $id,
+                code     => 504,
+                request  => $hashref->{request},
+                previous => $hashref->{previous},
+                content  => 'Timed out',
+            );
+
+            my $s = $hashref->{handle};
+            $self->_io_select->remove($s);
+            delete $$self{fileno_to_id}{ $s->fileno };
+        }
+    }
+
     return 1;
 }
 
@@ -598,6 +785,12 @@ sub _send_request {
 
     my %args = ();
 
+    # Get cookies from jar if one exists
+    my $jar = $self->_get_opt('cookie_jar', $id);
+    if ($jar) {
+        $jar->add_cookie_header($request);
+    }
+
     # We need to use a different request_uri for proxied requests. Decide to use
     # this if a proxy port or host is set.
     #
@@ -605,6 +798,14 @@ sub _send_request {
     $args{Host}     = $uri->host;
     $args{PeerAddr} = $self->_get_opt( 'proxy_host', $id );
     $args{PeerPort} = $self->_get_opt( 'proxy_port', $id );
+    $args{LocalAddr} = $self->_get_opt('local_addr', $id );
+    $args{LocalPort} = $self->_get_opt('local_port', $id );
+
+    # https://rt.cpan.org/Public/Bug/Display.html?id=33071
+    $args{Timeout} = $self->_get_opt( 'timeout', $id);
+
+    # ACF - Pass ssl_options through
+    $args{ssl_opts} = $self->_get_opt( 'ssl_options', $id);
 
     my $request_is_to_proxy =
       ( $args{PeerAddr} || $args{PeerPort} )    # if either are set...
@@ -612,10 +813,38 @@ sub _send_request {
       : 0;                                      # ...otherwise not
 
     # If we did not get a setting from the proxy then use the uri values.
+
     $args{PeerAddr} ||= $uri->host;
+    my $peer_address = $self->_get_opt('peer_addr', $id );
+    if($peer_address) {
+        $args{PeerAddr} = $peer_address;
+    }
+
     $args{PeerPort} ||= $uri->port;
 
-    my $s = eval { Net::HTTP::NB->new(%args) };
+    my $net_http_class = 'Net::HTTP::NB';
+    if ($uri->scheme and $uri->scheme eq 'https' and not $request_is_to_proxy) {
+        $net_http_class = 'Net::HTTPS::NB';
+        eval {
+            require Net::HTTPS::NB;
+            Net::HTTPS::NB->import();
+        };
+        die "$net_http_class must be installed for https support" if $@;
+
+        # Add SSL options, if any, to args
+        my $ssl_options = $self->_get_opt('ssl_options');
+        @args{ keys %$ssl_options } = values %$ssl_options if $ssl_options;
+    }
+    elsif($uri->scheme and $uri->scheme eq 'https' and $request_is_to_proxy) {
+        # We are making an HTTPS request through an HTTP proxy such as squid.
+        # The proxy will handle the HTTPS, we need to connect to it via HTTP
+        # and then make a request where the https is clear from the scheme...
+        $args{Host} = sprintf(
+            '%s:%s',
+            delete @args{'PeerAddr', 'PeerPort'}
+        );
+    }
+    my $s = eval { $net_http_class->new(%args) };
 
     # We could not create a request - fake up a 503 response with
     # error as content.
@@ -632,7 +861,10 @@ sub _send_request {
         return 1;
     }
 
-    my %headers = %{ $request->{_headers} };
+    my %headers;
+    for my $key ($request->{_headers}->header_field_names) {
+        $headers{$key} = $request->header($key);
+    }
 
     # Decide what to use as the request_uri
     my $request_uri = $request_is_to_proxy    # is this a proxy request....
@@ -645,16 +877,23 @@ sub _send_request {
 
     $self->_io_select->add($s);
 
-    $$self{fileno_to_id}{ $s->fileno }   = $id;
-    $$self{in_progress}{$id}{request}    = $request;
-    $$self{in_progress}{$id}{timeout_at} =
-      time + $self->_get_opt( 'timeout', $id );
-    $$self{in_progress}{$id}{finish_by} =
-      time + $self->_get_opt( 'max_request_time', $id );
+    my $time = time;
+    my $entry = $$self{in_progress}{$id} ||= {};
 
-    $$self{in_progress}{$id}{redirects_left} =
-      $self->_get_opt( 'max_redirects', $id )
-      unless exists $$self{in_progress}{$id}{redirects_left};
+    $$self{fileno_to_id}{ $s->fileno } = $id;
+
+    $entry->{request}    = $request;
+    $entry->{started_at} = $time;
+
+    
+    $entry->{timeout_at} = $time + $self->_get_opt( 'timeout', $id );
+    # warn "sent - timeout set to '$entry->{timeout_at}'";
+
+    $entry->{finish_by}  = $time + $self->_get_opt( 'max_request_time', $id );
+    $entry->{handle}     = $s;
+
+    $entry->{redirects_left} = $self->_get_opt( 'max_redirect', $id )
+      unless exists $entry->{redirects_left};
 
     return 1;
 }
@@ -682,16 +921,7 @@ sub _make_url_absolute {
     my $in  = $args{url};
     my $ref = $args{ref};
 
-    return $in if $in =~ m{ \A http:// }xms;
-
-    my $ret = $ref->scheme . '://' . $ref->authority;
-    return $ret . $in if $in =~ m{ \A / }xms;
-
-    $ret .= $ref->path;
-    return $ret . $in if $in =~ m{ \A [\?\#\;] }xms;
-
-    $ret =~ s{ [^/]+ \z }{}xms;
-    return $ret . $in;
+    return URI->new_abs($in, $ref)->as_string;
 }
 
 sub _add_error_response_to_return {
@@ -722,6 +952,7 @@ by domain so that the remote server is not overloaded.
 =head1 GOTCHAS
 
 The responses may not come back in the same order as the requests were made.
+For https requests to work, you must have L<Net::HTTPS::NB> installed.
 
 =head1 THANKS
 
@@ -731,6 +962,30 @@ before headers sent and more.
 Tomohiro Ikebe from livedoor.jp submitted patches (and a test) to properly
 handle 304 responses.
 
+Naveed Massjouni for adding the https handling code.
+
+Alex Balhatchet for adding the https + proxy handling code, and for making the
+tests run ok in parallel.
+
+Josef Toman for fixing two bugs, one related to header handling and another
+related to producing an absolute URL correctly.
+
+Github user 'c00ler-' for adding LocalAddr and LocalPort support.
+
+rt.cpan.org user 'Florian (fschlich)' for typo in documentation.
+
+Heikki Vatiainen for the ssl_options support patch.
+
+Daniel Lintott of the Debian Perl Group for pointing out a test failure when
+using a very recent version of HTTP::Server::Simple to implement
+t/TestServer.pm
+
+=head1 BUGS AND REPO
+
+Please submit all bugs, patches etc on github
+
+L<https://github.com/evdb/HTTP-Async>
+
 =head1 AUTHOR
 
 Edmund von der Burg C<< <evdb@ecclestoad.co.uk> >>. 
